/*
** dtools.c - utility functions used by voiceover server code
** Copyright (C) 2019, 2020  Lew Pitcher
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License along
** with this program; if not, write to the Free Software Foundation, Inc.,
** 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
**
** Lew Pitcher <lew.pitcher@digitalfreehold.ca>
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pwd.h>
#include <syslog.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>

#include "dtools.h"

/*
** Daemonize() will convert the process into a daemon
** by forking() off and orphaning a child process
** (so that it becomes a process that's not a process group leader)
** The orphan child process then uses setsid() to become
** a session leader in a new session, and the process group
** leader of a new process group within that session.
** It will not have a controlling terminal (courtesy of setsid()).
** Daemonize() will then close all FDs except for stdin, stdout,
** and stderr. If save_stdio is zero, Daemonize will redirect
** stdin, stdout and stderr to /dev/null, otherwise it will leave
** them unchanged.
**
** Failures, if any, will log to syslog
*/
int Daemonize(int save_stdio)
{
  char	*rootdir = "/";

  int rc = DAEMONIZED_ERR;

  switch(fork())
  {
    case 0:		/* fork success - we are child */
      if (setsid() > 0)
      {
	if (chdir(rootdir) == 0)
	{
	  struct rlimit maxfds;
	  int fd;

	  umask(0);

	  if (save_stdio == DAEMONIZE_NULL_STDIO)
	  {
	    close(STDIN_FILENO);
	    close(STDOUT_FILENO);
	    close(STDERR_FILENO);	
	    if ((fd = open("/dev/null",O_RDWR)) == STDIN_FILENO)
	    {
	      dup(fd);
	      dup(fd);
	    }
	    else close(fd);
	  }

	  getrlimit(RLIMIT_NOFILE,&maxfds);
	  for (fd = 3; fd < maxfds.rlim_max; ++fd) close(fd);

	  rc = DAEMONIZED_OK;
	}
	else syslog(LOG_CRIT,"Cant chdir to %s - errno = %d (%s)",rootdir,errno,strerror(errno));
      }
      else syslog(LOG_CRIT,"Cant occupy new session - errno = %d (%s)",errno,strerror(errno));
      break;

    case -1:
      syslog(LOG_CRIT,"Cant fork - errno = %d (%s)",errno,strerror(errno));
      break;

    default:		/* fork success - we are parent */
      /*
	there's nothing more the parent process can, or should do
	it just needs to silently vanish away
      */
      _exit(0);
  }
  return rc;
}

/*
** ChangeUser will cause the process to assume the euid and egid
** of the named user. If NULL presented as the name, then ChangeUser
** will attempt to reset euid to ruid and egid to rgid.
**
** Failures, if any, will log to syslog
*/
int ChangeUser(char *newuser)
{
  int rc = CHANGEUSER_ERR;	/* failure */

  if (getuid() == 0)
  {
    if (newuser)	/* change euid/egid to specified user */
    {
      struct passwd *user;
      if ((user = getpwnam(newuser)))
      {
	if ((getegid() != user->pw_gid) && setegid(user->pw_gid))
	  syslog(LOG_CRIT,"Cant change real GID to %lu",(unsigned long)user->pw_gid);
	else
	{
	  if ((geteuid() != user->pw_uid) && seteuid(user->pw_uid))
	    syslog(LOG_CRIT,"Cant change real UID to %lu",(unsigned long)user->pw_uid);
	  else
	    rc = CHANGEUSER_OK;	/* everything worked */
	}
      }
      else syslog(LOG_CRIT,"Cant change user: No user named %s in passwd file",newuser);
    }
    else /* restore euid/egid to ruid/rgid */
    {
      if (seteuid(getuid()) == 0)
	if (setegid(getgid()) == 0)
	  rc = CHANGEUSER_OK;
	else
	  syslog(LOG_CRIT,"Cant change effective gid back to gid %lu - %s",(unsigned long)getgid(),strerror(errno));
      else
	syslog(LOG_CRIT,"Cant change effective uid back to uid %lu - %s",(unsigned long)getuid(),strerror(errno));
    }
  }
  else syslog(LOG_ERR,"To change user, this program must run as root");
  return rc;
}

/*
** PidFile writes the PID of the current process out to the named file
**
** Failures, if any, will log to syslog
*/
int PidFile(char *fpath)
{
  int rc = 0;	/* 0 is failure, 1 is success */
  FILE *pidfile;
  mode_t old_umask;

  old_umask = umask(0133);

  if ((pidfile = fopen(fpath,"w")))
  {
    fprintf(pidfile,"%lu\n",(unsigned long)getpid());
    fclose(pidfile);
    rc = 1;
  }
  else syslog(LOG_INFO,"Failed to open PIDFILE %s for writing - no PIDFILE written",fpath);

  umask(old_umask);

  return rc;
}
