/*
** flite.c - Interface to the CMU Flite tts system from voxd VoiceOver daemon
** Copyright (C) 2017, 2020  Lew Pitcher
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License along
** with this program; if not, write to the Free Software Foundation, Inc.,
** 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
**
** Lew Pitcher <lew.pitcher@digitalfreehold.ca>
*/

/*
** NOTE: This version REQUIRES libraries from flite v2.0.0
**	 This version REQUIRES pulseaudio
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <syslog.h>

#include <pulse/simple.h>
#include <pulse/error.h>

#include <flite/flite.h>

#define PULSE_APP_NAME	"VoiceOver"
#define	PULSE_STREAM	"TTS"

cst_voice *register_cmu_us_rms(char *);	/* libflite_cmu_us_rms */
void usenglish_init(cst_voice *v);	/* libflite_usenglish */
cst_lexicon *cmulex_init(void);		/* libflite_cmulex */

static pa_sample_spec ss = { .format = PA_SAMPLE_S16NE, .channels = 1, .rate = 16000 };
static pa_simple *s = NULL;		/* pulseaudio simple handle */

static int PulseFlite(const cst_wave *, int, int, int, cst_audio_streaming_info *);

int InitFlite(void)
{
  int rc = 0;		/* FAILURE */

  s = pa_simple_new(NULL,		// Use the default server.
                  PULSE_APP_NAME,	// Our application's name.
                  PA_STREAM_PLAYBACK,	// We only play back audio
                  NULL,			// Use the default device.
                  PULSE_STREAM,		// Description of our stream.
                  &ss,			// Our sample format.
                  NULL,			// Use default channel map
                  NULL,			// Use default buffering attributes.
                  NULL			// Ignore error code.
                  );
  if (s)
  {
    flite_init();

    flite_add_lang("eng",usenglish_init,cmulex_init);
    flite_add_lang("usenglish",usenglish_init,cmulex_init);

    flite_voice_list = cons_val(voice_val(register_cmu_us_rms(NULL)),flite_voice_list);
    flite_voice_list = val_reverse(flite_voice_list);
    rc = 1;		/* SUCCESS */
  }
  else syslog(LOG_ERR,"pa_simple_new() couldnt establish our pulseaudio connection");

  return rc;
}

cst_voice *GetVoice(char *path, float tempo, float pitch, float intonation)
{
  cst_voice *Voice = NULL;

  if ((Voice = flite_voice_select(path)))
  {
    cst_audio_streaming_info *asi;

    /* set up for pulseaudio playback callback */
    asi = new_audio_streaming_info();
    asi->asc = PulseFlite;
    asi->userdata = s;

    feat_set(Voice->features,"streaming_info",audio_streaming_info_val(asi));

    if (tempo > 0.0)
      feat_set_float(Voice->features,"duration_stretch",tempo);
    if (pitch > 0.0)
      feat_set_float(Voice->features,"int_f0_target_mean",pitch);
    if (intonation > 0.0)
      feat_set_float(Voice->features,"int_f0_target_stddev",intonation);
  }
  return Voice;
}

void Say(cst_voice *Voice, char *format, ...)
{
  va_list	ap, 	/* our va_list */
		pp; 	/* vsnprintf va_list */

  size_t	textsize;
  char		dummy[1],
		*text = NULL;

  va_start(ap,format);

  va_copy(pp,ap);
  textsize = 2 + vsnprintf(dummy,sizeof(dummy),format,pp);

  if ((text = malloc(textsize)))
  {
    va_copy(pp,ap);
    vsnprintf(text,textsize,format,pp);
    flite_text_to_speech(text,Voice,"none");
    free(text);
  }
  va_end(ap);
}

void TermFlite(void)
{
  int pa_error;

  /* wait till all audio externalized */
  if (pa_simple_drain(s,&pa_error))
    syslog(LOG_ERR,"pa_simple_drain() returned error %d (\"%s\")\n",pa_error,pa_strerror(pa_error));

  /* close the pulseaudio feed */
  pa_simple_free(s); s = NULL;

  /* NB: flite doesn't need a cleanup */
}

/*
** PulseFlite - flite callback to pulseaudio to play audio segment
*/
static int PulseFlite(const cst_wave *w, int start, int size, int last, cst_audio_streaming_info *asi)
{
  if (size > 0) /* then we have something to say */
  {
    int error, pa_error;
    error = pa_simple_write(asi->userdata,		/* our PulseAudio handle */
			    &w->samples[start],		/* start of data */
			    size*sizeof(short),		/* length in bytes */
			    &pa_error);			/* place to store error code */

    /* if error is non-zero, then pa_errno is the pulseaudio error number */
    if (error)
      syslog(LOG_ERR,"pa_simple_write() returned error %d (\"%s\")\n",pa_error,pa_strerror(pa_error));
  }

  /* <<last>> is true on last segment, otherwise false */
  return last ? CST_AUDIO_STREAM_STOP : CST_AUDIO_STREAM_CONT;
}
