%{
  /* voxd.y - YACC grammar parser for voxd VoiceOver daemon config files
  ** Copyright (C) 2017, 2020  Lew Pitcher
  **
  ** This program is free software; you can redistribute it and/or modify
  ** it under the terms of the GNU General Public License as published by
  ** the Free Software Foundation; either version 2 of the License, or
  ** (at your option) any later version.
  **
  ** This program is distributed in the hope that it will be useful,
  ** but WITHOUT ANY WARRANTY; without even the implied warranty of
  ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  ** GNU General Public License for more details.
  **
  ** You should have received a copy of the GNU General Public License along
  ** with this program; if not, write to the Free Software Foundation, Inc.,
  ** 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  **
  ** Lew Pitcher <lew.pitcher@@digitalfreehold.ca>
  */

  /* headers, globals, etc */
  #include <stdio.h>
  #include <stdlib.h>
  #include <unistd.h>
  #include <string.h>
  #include <stdarg.h>
  #include <syslog.h>
  #include <sys/types.h>
  #include <sys/stat.h>

  #include "voxd.h"

  extern int yylex(void);

  static struct show	*newShow(void);
  static void		setShowAct(struct show *, int, char *);
  static void		delShow(struct show *);

  static struct actor	*newActor(void);
  static void		setActorVoiceParm(struct actor *, int, float);
  static void		delActors(struct actor *);
  static struct actor	*findActor(struct actor *, char *);

  static struct role	*newRole(void);
  static struct role	*findRole(struct role *, char *);
  static void		delRoles(struct role *);

  void		yyerror(char *,...);

  unsigned int	ErrorCount = 0,
		LineNumber = 1;
  struct actor	*CastList = NULL;
  struct show	*AudioPlay = NULL;


%}
%token PROLOGUE EPILOGUE REFRESH VOICEDBY HASVOICE LPAREN RPAREN EQUALS COMMA TEMPO PITCH INTONATION
%union
{
  char		*String;
  float		Float;
  int		Number;
  struct actor	*Actor;
  struct role   *Role;
  struct show	*Show;
}
%token <String> QUOTEDSTRING NAME
%token <Float>  FLOATPT
%type  <Number> parm mark
%type  <Actor>	mods voice actor actors
%type  <Role>	role roles
%type  <Show>	act acts show
%%
vox	: show
	{
	  struct show *show = $1;

	  if (ErrorCount)
	  {
	    fprintf(stderr,"%d Errors found\n",ErrorCount);
	    delShow(show);
	    show = NULL;
	  }
	  else
	  {
	    /* ensure that we have a "CHORUS" role and actor */
	    if (findRole(show->roles,CHORUS_ROLE) == NULL)
	    {
	      struct actor *a_chorus;
	      struct role *r_chorus;

	      if ((a_chorus = newActor()))
	      {
		if ((r_chorus = newRole()))
		{
		  r_chorus->roleName  = strdup(CHORUS_ROLE);
		  r_chorus->roleActor = a_chorus;
		  a_chorus->actorName = strdup(CHORUS_ACTOR);
		  a_chorus->voicePath = strdup(CHORUS_VOICE);

		  r_chorus->roleNext  = show->roles;	show->roles  = r_chorus;
		  a_chorus->actorNext = show->actors;	show->actors = a_chorus;
		}
		else
		{
		  free(a_chorus);
		  yyerror("cant allocate space for CHORUS role");
		}
	      }
	      else yyerror("cant allocate space for CHORUS actor");
	    }
	    if (ErrorCount)
	    {
	      delShow(show);
	      show = NULL;
	    }
	  }
	  AudioPlay = show;
	}
	;

show	: acts actors roles
	{
	  struct show *show = $1;
	  if (show)
	  {
	    show->actors = $2;
	    show->roles  = $3;
	  }
	  $$ = show;
	}
	| actors roles
	{
	  struct show *show = newShow();

	  if (show)
	  {
	    show->actors = $1;
	    show->roles  = $2;
	  }
	  $$ = show;
	}
	;

acts	: acts act
	{
	  struct show	*show = $1,
			*act  = $2;

	  if (act->prologue)
	  {
	    free(show->prologue);
	    show->prologue = act->prologue;
	  }
	  else
	    if (act->epilogue)
	    {
	      free(show->epilogue);
	      show->epilogue = act->epilogue;
	    }
	    else
	    {
	      free(show->refresh);
	      show->refresh = act->refresh;
	    }
	  free(act);
	  $$ = show;
	}
	| act
	{
	  $$ = $1;
	}
	;

act	: mark EQUALS QUOTEDSTRING
	{
	  struct show *show = newShow();
	  setShowAct(show,$1,$3);
	  $$ = show;
	}
	;

mark	: PROLOGUE
	{ $$ = PROLOGUE; }
	| EPILOGUE
	{ $$ = EPILOGUE; }
	| REFRESH
	{ $$ = REFRESH; }
	;

actors	: actors actor
	{
	  struct actor *actor = $2;

	  if (actor)
	    actor->actorNext = $1;
	  else
	    actor = $1;

	  $$ = CastList = actor;
	}
	| actor
	{
	  $$ = CastList = $1;
	}
	| error
	{
	  $$ = NULL;
	}
	;

actor	: NAME hasvoice voice
	{
	  struct stat buf;
	  struct actor *actor;

	  actor = $3;
	  if (actor)
	  {
	    actor->actorName = $1;
	    if (stat(actor->voicePath,&buf) || !S_ISREG(buf.st_mode))
	      yyerror("Actor %s: invalid voice %s",actor->actorName,actor->voicePath);
	  }

	  $$ = actor;
	}
	;

hasvoice : HASVOICE
	 | EQUALS
	 ;

roles	: roles role
	{
	  struct role *role = $2;
	  if (role)
	    role->roleNext = $1;
	  else
	    role = $1;
	  $$ = role;
	}
	| role
	{
	  $$ = $1;
	}
	;

role	: NAME voicedby NAME
	{
	  struct role *role = newRole();
	  if (role)
	  {
	    if ((role->roleActor = findActor(CastList,$3)) == NULL)
	    {
	      yyerror("For role %s, actor %s has not been defined",$1,$3);
	      free(role); role = NULL;
	    }
	    else role->roleName = $1;
	  }
	  $$ = role;
	}
	| error
	{
	  $$ = NULL;
	}
	;

voicedby : VOICEDBY
	 | EQUALS
	 ;

voice	: QUOTEDSTRING LPAREN mods RPAREN
	{
	  struct actor *actor;

	  actor = $3;
	  if (actor) actor->voicePath = $1;
	  $$ = actor;
	}
	| QUOTEDSTRING
	{
	  struct actor *actor;

	  actor = newActor();
	  if (actor) actor->voicePath = $1;
	  $$ = actor;
	}
	;

mods	: mods COMMA parm EQUALS FLOATPT
	{
	  setActorVoiceParm($1,$3,$5);
	  $$ = $1;
	}
	| parm EQUALS FLOATPT
	{
	  struct actor *actor;

	  actor = newActor();
	  setActorVoiceParm(actor,$1,$3);
	  $$ = actor;
	}
	;

parm	: TEMPO
	{ $$ = TEMPO; }
	| PITCH
	{ $$ = PITCH; }
	| INTONATION
	{ $$ = INTONATION; }
	;
%%
void yyerror(char *msg,...)
{
  va_list ap;

#ifdef LOG_TO_STDERR
  fprintf(stderr, "configuration error near line %d: ",LineNumber);
  va_start(ap,msg); vfprintf(stderr,msg,ap); va_end(ap);
  fputc('\n',stderr);
#else
  syslog(LOG_ERR,"configuration error near line %d: ",LineNumber);
  va_start(ap,msg); vsyslog(LOG_ERR,msg,ap); va_end(ap);
#endif

  ++ErrorCount;
}

int yywrap()
{
  return 1;
}

/*
**************************************************************************************
*/

static struct show *newShow(void)
{
  struct show *this;

  if ((this = malloc(sizeof *this)))
  {
    this->prologue = NULL;
    this->epilogue = NULL;
    this->refresh  = NULL;
    this->actors   = NULL;
    this->roles    = NULL;
  }
  else yyerror("Cannot allocate the show");

  return this;
}

static void delShow(struct show *show)
{
  if (show)
  {
    delRoles(show->roles);
    delActors(show->actors);

    free(show->prologue);
    free(show->epilogue);
    free(show->refresh);

    free(show);
  }
}

static void setShowAct(struct show *show, int act, char *line)
{
  if (show)
  {
    switch (act)
    {
      case PROLOGUE:
	free(show->prologue);
	show->prologue = line;
	break;

      case EPILOGUE:
	free(show->epilogue);
	show->epilogue = line;
	break;

      case REFRESH:
	free(show->refresh);
	show->refresh = line;
	break;
    }
  }
}

static struct role *newRole(void)
{
  struct role *this;

  if ((this = malloc(sizeof *this)))
  {
    this->roleNext  = NULL;
    this->roleName  = NULL;
    this->roleActor = NULL;
  }
  else yyerror("Cannot allocate a role");

  return this;
}

static struct role *findRole(struct role *list, char *name)
{
  struct role *role = NULL;

  for (role = list; role ; role = role->roleNext)
    if (strcmp(role->roleName,name) == 0) break;

  return role;
}

static void delRoles(struct role *list)
{
  if (list)
  {
    delRoles(list->roleNext);
    free(list->roleName);
    free(list);
  }
}

static struct actor *findActor(struct actor *list, char *name)
{
  struct actor *here;

  for (here = list; here; here = here->actorNext)
    if (strcmp(name,here->actorName) == 0) break;
  return here;
}

static void setActorVoiceParm(struct actor *this, int parm, float valu)
{
  if (this)
  {
    switch (parm)
    {
      case TEMPO:
	this->voiceTempo = valu;
	break;

      case PITCH:
	this->voicePitch = valu;
	break;

      case INTONATION:
	this->voiceIntonation = valu;
	break;
    }
  }
}

static void delActors(struct actor *list)
{
  if (list)
  {
    delActors(list->actorNext);
    free(list->actorName);
    free(list->voicePath);
    free(list);
  }
}

static struct actor *newActor(void)
{
  struct actor *this;

  if ((this = malloc(sizeof *this)))
  {
    this->actorNext	  = NULL;
    this->actorName	  = NULL;
    this->voicePath	  = NULL;
    this->voiceTempo	  = 0.0;
    this->voicePitch	  = 0.0;
    this->voiceIntonation = 0.0;
  }
  else yyerror("Cannot allocate an actor");

  return this;
}

/*
**************************************************************************************
*/

/*
** OpenShow()	Construct VoiceOver configuration tree
** Accepts:	pointer to character string naming configuration file path
** Process:	read config file to build config tree
** Returns:	pointer to the root of config tree
*/
struct show *OpenShow(char *configpath)
{
  extern FILE *yyin;
  struct show *rv = NULL;

  ErrorCount	= 0;
  LineNumber	= 1;
  CastList	= NULL;
  AudioPlay	= NULL;

  if ((yyin = fopen(configpath,"r")))
  {
    if ((yyparse() == 0) && (ErrorCount == 0))
      rv = AudioPlay;
    fclose(yyin);
  }
  else perror(configpath);

  return rv;  
}

/*
** CloseShow()	Demolish VoiceOver configuration tree
** Accepts:	pointer to the root of config tree
** Process:	release all storage allocated to the tree
** Returns:	nothing
*/
void CloseShow(struct show *show)
{
  delShow(show);
}

/*
** WriteShow()	Externalize VoiceOver configuration tree
** Accepts:	FILE pointer to open stdio file,
**		pointer to the root of the config tree
** Process:	writes textual representation of VoiceOver
**		configuration tree to specified file
** Returns:	nothing
*/
void WriteShow(FILE *fp, struct show *show)
{
  struct actor *actor;
  struct role *role;

  if ((fp == NULL) || (show == NULL)) return;

  if (show->prologue || show->epilogue || show->refresh)
  {
    char *format = "%s = \"%s\"\t; This spoken %s\n";

    fputs("; voxd will announce itself\n",fp);
    if (show->prologue)
      fprintf(fp,format,"Prologue",show->prologue,"at startup");

    if (show->epilogue)
      fprintf(fp,format,"Epilogue",show->epilogue,"at shutdown");

    if (show->refresh)
      fprintf(fp,format,"NewCast",show->refresh,"after config refresh");

    fputc('\n',fp);
  }

  fputs("; The ACTORS and their voices\n",fp);
  for (actor = show->actors; actor; actor = actor->actorNext)
  {
    /* Generate a comment to document the internally-generated CHORUS_ACTOR */
    if (strcmp(actor->actorName,CHORUS_ACTOR) == 0) fputs("; ",fp);

    fprintf(fp,"%s has voice \"%s\"",actor->actorName,actor->voicePath);
    if (actor->voiceTempo || actor->voicePitch || actor->voiceIntonation)
    {
      char *comma = "";

      fputs(" (",fp);

      if (actor->voiceTempo)
      {
	fprintf(fp,"tempo=%-.3f",actor->voiceTempo); comma = ", ";
      }

      if (actor->voicePitch)
      {
	fprintf(fp,"%spitch=%-.3f",comma,actor->voicePitch); comma = ", ";
      }

      if (actor->voiceIntonation) fprintf(fp,"%sintonation=%-.3f",comma,actor->voiceIntonation);
      fputc(')',fp);
    }
    fputc('\n',fp);
  }
  fputc('\n',fp);

  fputs("; The ROLES and their ACTORS\n",fp);
  for (role = show->roles; role; role = role->roleNext)
  {
    /* Generate a comment to document the internally-generated CHORUS_ROLE */
    if ((strcmp(CHORUS_ROLE,role->roleName) == 0) && (strcmp(CHORUS_ACTOR,role->roleActor->actorName) == 0))
      fputs("; ",fp);

    fprintf(fp,"%s is voiced by %s\n",role->roleName, role->roleActor->actorName);
  }
}

/*
** GetActorInRole	Retrieve voice particulars for a specified role
** Accepts:	pointer to character string naming the role to retrieve
**		pointer to struct show of root of config tree
** Process:	look through config tree for role
**		if found, return the actor for that role
**		otherwise, look through config tree for actor
**		if found, return the actor /named/ by the role
**		othewise, return the actor that voices for CHORUS
** Returns:	pointer to struct actor of the voice for the specified role
*/
struct actor *GetActorInRole(char *Role, struct show *Show)
{
  struct actor	*actor = NULL,
		*chorus = NULL;
  struct role	*role;

  for (role = Show->roles; role; role = role->roleNext)
  {
    if (strcmp(CHORUS_ROLE,role->roleName) == 0)	chorus = role->roleActor;
    if (strcmp(Role,role->roleName) == 0)		actor  = role->roleActor;
  }

  if (actor == NULL)	/* No actor for this role, perhaps role /is/ actor? */
    for (actor = Show->actors; actor && (strcmp(actor->actorName,Role) != 0); actor = actor->actorNext)
      continue;	/* search list of actors for an actor who's name matches the requested Role */

  if (actor == NULL)	/* couldn't find the proper actor */ 
    actor = chorus;	/* so use the chorus instead	  */

  return actor;
}
