/*
** VoiceOver - a speech serializer for flite
**
** voxd_main	mainline for voxd VoiceOver daemon
** Copyright (C) 2017, 2018, 2019, 2020  Lew Pitcher
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License along
** with this program; if not, write to the Free Software Foundation, Inc.,
** 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
**
** Lew Pitcher <lew.pitcher@digitalfreehold.ca>
*/

#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <syslog.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/select.h>

#include "voxconfig.h"
#include "voxd.h"
#include "flite.h"
#include "dtools.h"

/*
**  ---------- LOCAL FUNCTION PROTOTYPES ----------
*/
static int  VoiceOver(char *, int);				/* main processing routine - run in daemon */
static void ReadLine(struct show *, char *, char *);		/* speaks text out loud */

static void RequestStop(int);			/* interrupt handler - clears Processing flag */
static void RefreshConfig(int);			/* interrupt handler - sets ReloadConfig flag */
static int  GetArgs(int , char **, char **, int *, char **, char **, char **, char **);	/* cmdline argument parser */
static char *MakePath(char *);			/* convert given path into fully qualified pathname */

/*
**  ---------- MACROS and MANIFEST CONSTANTS ----------
*/

/* for select() call		*/
#define	WAIT_SECS		15
#define	WAIT_USEC		0

/* PAUSE this long (in seconds) between role changes */
#define	PAUSE_SECS		2

/*
**  ---------- GLOBAL VARIABLES ----------
*/
static volatile int Processing = 1;		/* normally 1, set to 0 when termination requested */
static volatile int ReloadConfig  = 0;		/* normally 0, set to 1 when config file reload requested */

/*
**  ---------- PRINCIPAL LOGIC ----------
*/
int main(int argc, char *argv[])
{
  int	exit_status = EXIT_FAILURE;
  char	*ConfigPath	= NULL,
	*LockFilePath	= NULL,
	*PidFilePath	= NULL,
	*SocketPath	= NULL,
	*UserName	= NULL;
  int	runInForeground;

  openlog(argv[0],LOG_PID|LOG_PERROR,LOG_DAEMON);

  if (GetArgs(argc,argv,&ConfigPath,&runInForeground,&LockFilePath,&PidFilePath,&SocketPath,&UserName) == 0)
  {
    int lockfile;

    if ((lockfile = open(LockFilePath,O_WRONLY|O_CREAT|O_EXCL,0600)) >= 0)
    {
      if (flock(lockfile,LOCK_EX|LOCK_NB) == 0)
      {
	if (runInForeground == 1 || Daemonize(DAEMONIZE_NULL_STDIO) == DAEMONIZED_OK)
	{
	  int script;

	  /* open socket */
	  if ((script = socket(AF_UNIX,SOCK_DGRAM,0)) >= 0)
	  {
	    struct sockaddr_un serveraddress;

	    umask(0155);
	    unlink(SocketPath);
	    memset(&serveraddress,0,sizeof serveraddress);
	    serveraddress.sun_family = AF_LOCAL;
	    strcpy(serveraddress.sun_path,SocketPath);
	    if (bind(script,(struct sockaddr *)&serveraddress,sizeof serveraddress) == 0)
	    {
	      PidFile(PidFilePath);
	      if ((UserName == NULL) || (ChangeUser(UserName) == CHANGEUSER_OK))
	      {
		if (VoiceOver(ConfigPath,script)) exit_status = EXIT_SUCCESS;

		/* if we changed user to get here, change back before we leave */
		if (UserName) ChangeUser(NULL);
	      }
	      unlink(PidFilePath);
	    }
	    else  syslog(LOG_CRIT,"Failed to bind socket to %s (%s) - aborting",SocketPath,strerror(errno));
	    close(script);
	    unlink(SocketPath);
	  }
	  else syslog(LOG_CRIT,"Failed to create UNIX DATAGRAM socket (%s) - aborting",strerror(errno));
	}
	else syslog(LOG_CRIT,"Cant daemonize - aborting");
      }
      else  syslog(LOG_CRIT,"Cant acquire exclusive lock on %s (%s) - aborting",LockFilePath,strerror(errno));

      close(lockfile);
      unlink(LockFilePath);
    }
    else syslog(LOG_CRIT,"Cant allocate lockfile %s (%s) - aborting",LockFilePath,strerror(errno));
  }
  else syslog(LOG_CRIT,"Invalid commandline options- aborting");
  closelog();

  return exit_status;
}

static int VoiceOver(char *ConfigPath, int script)
{
  int	rc = 0;		/* FAILURE */
  struct show *Show;

  if ((Show = OpenShow(ConfigPath)))
  {
    if (InitFlite())	/* no problems initializing tts system */
    {
      char	*prevRole = NULL;
      time_t	timePrev = 0,
		timeDiff;

      /* establish signal handlers to set the refresh-config and stop-looping flags */
      signal(SIGHUP, RefreshConfig);
      signal(SIGINT, RequestStop);
      signal(SIGTERM,RequestStop);

      if (Show->prologue) ReadLine(Show,"",Show->prologue);
      while (Processing)
      {
	fd_set rfds;
	struct timeval tv;

	char	buffer[MAXMSGLEN+1],
		*Line;
	ssize_t	msglen;

	if (ReloadConfig)
	{
	  struct show *NewShow;

	  if ((NewShow = OpenShow(ConfigPath)))
	  {
	    syslog(LOG_INFO,"Got new Show from %s",ConfigPath);
	    CloseShow(Show);
	    Show = NewShow;
	    if (Show->refresh) ReadLine(Show,"",Show->refresh);
	  }
	  ReloadConfig = 0;
	}

	/* wait until line ready or timer goes off */
	FD_ZERO(&rfds); FD_SET(script,&rfds);
	tv.tv_sec = WAIT_SECS; tv.tv_usec = WAIT_USEC;
	if (select(script+1,&rfds,NULL,NULL,&tv) > 0)
	{
	  timeDiff = time(NULL) - timePrev;

	  msglen = read(script,buffer,(sizeof buffer)-1);
	  buffer[msglen] = 0;
	  if (msglen > 0 && (Line = strchr(buffer,':')) != NULL)
	  {
	    *Line++  = 0;	/* make first part of buffer into role name string */
	    while (isspace(*Line)) ++Line;	/* skip leading whitespace in Line */

	    /* When the role changes, we insert a pause to reduce listener confusion */
	    if (prevRole && strcmp(prevRole,buffer) != 0 && timeDiff < PAUSE_SECS)
	      sleep(PAUSE_SECS - timeDiff);
	    free(prevRole); prevRole = strdup(buffer);

	    ReadLine(Show,buffer,Line);
	    timePrev = time(NULL);
	  }
	}
      }
      if (Show->epilogue) ReadLine(Show,"",Show->epilogue);
      TermFlite();
      rc = 1;	/* SUCCESS */
    }
    CloseShow(Show);
  }
  else syslog(LOG_CRIT,"Failed to obtain configuration from %s - aborting",ConfigPath);

  return rc;
}

static void ReadLine(struct show *Cast, char *Role, char *Line)
{
  struct actor *Actor;

  if ((Actor = GetActorInRole(Role,Cast)))
  {
    cst_voice *Voice;

    if ((Voice = GetVoice(Actor->voicePath,Actor->voiceTempo,Actor->voicePitch,Actor->voiceIntonation)))
      Say(Voice,"%s",Line);
    else
      syslog(LOG_ERR,"Actor %s has lost their voice and cant say the line",Actor->actorName);
  }
  else syslog(LOG_ERR,"Role %s has no actor and the line remains unsaid",Role);
}

/*
**  ---------- SUPPORT FUNCTIONS ----------
*/

/*
** configuration that can be altered from the commandline:
** - the path to the config file
**	-c <configpath>
** - the username under which to run
**	-u username
** - whether or not the program will daemonize
**	-d
** - the path to the lockfile
**	-l <lockfilepath>
** - the path to the pidfile
**	-p <pidfilepath>
** - the path to the unix socket
**	-s <socketfilepath>
*/
static int GetArgs(int argc, char *argv[],
	    char **configpath,
	    int *foreground,
	    char **lockfilepath,
	    char **pidfilepath,
	    char **socketpath,
	    char **username)
{
  int	opt;
  extern char *optarg;

  char	*ConfigPath	= NULL,
	*LockFilePath	= NULL,
	*PidFilePath	= NULL,
	*SocketPath	= NULL,
	*UserName	= NULL;
  int	runInForeground = 1,
	argerrors	= 0;

  while ((opt = getopt(argc,argv,"c:u:l:p:s:d")) != -1)
  {
    switch (opt)
    {
      case 'c':	/* change configfile path */
	free(ConfigPath);
	if ((ConfigPath = realpath(optarg,NULL)) == NULL)
	{
	  fprintf(stderr,"%s: Cannot resolve %s into a full config file path\n",argv[0],optarg);
	  argerrors = 1;
	}
	break;

      case 'u': /* change execution user/group */
	free(UserName);
	UserName = strdup(optarg);
	break;

      case 'd':	/* run as daemon */
	runInForeground = 0;
	break;

      case 'l':	/* change lockfile path */
	free(LockFilePath);
	if ((LockFilePath = MakePath(optarg)) == NULL)
	{
	  fprintf(stderr,"%s: Encountered a problem expanding %s into a full lockfile path\n",argv[0],optarg);
	  argerrors = 1;
	}
	break;

      case 'p':	/* change pidfile path */
	free(PidFilePath);
	if ((PidFilePath = MakePath(optarg)) == NULL)
	{
	  fprintf(stderr,"%s: Encountered a problem expanding %s into a full PID file path\n",argv[0],optarg);
	  argerrors = 1;
	}
	break;

      case 's':	/* change socket path */
	free(SocketPath);
	if ((SocketPath = MakePath(optarg)) == NULL)
	{
	  fprintf(stderr,"%s: Encountered a problem expanding %s into a full socket path\n",argv[0],optarg);
	  argerrors = 1;
	}
	break;

      default:	/* anything else */
	argerrors = 1;
	break;
    }
  }

  if (argerrors == 0)
  {
    *configpath   = ConfigPath   ? ConfigPath   : strdup(CONFIGPATH);
    *foreground   = runInForeground;
    *lockfilepath = LockFilePath ? LockFilePath : strdup(LOCKFILEPATH);
    *pidfilepath  = PidFilePath  ? PidFilePath  : strdup(PIDFILEPATH);
    *socketpath   = SocketPath   ? SocketPath   : strdup(SOCKETPATH);
    *username	  = UserName;
  }
  else fprintf(stderr,"Usage: %s [-c configpath] [-u username] [-d] [-l lockfilepath] [-p pidfilepath] [-s socketpath]\n",argv[0]);

  return argerrors;
}

/*
** convert given (possibly relative) path into
** fully qualified pathname
** Returns pointer to malloc()ed full path name
** or NULL (on error)
*/
static char *MakePath(char *partpath)
{
  char *fullpath = NULL;

  if (partpath && *partpath)
  {
    if (*partpath != '/')
    {
      char path[1+(2 * PATH_MAX)];

      if ((getcwd(path, sizeof path)) && (path[0] == '/'))
      {
	if ((fullpath = malloc(strlen(path)+1+strlen(partpath)+1)))
	  sprintf(fullpath,"%s/%s",path,partpath);
      }
    }
    else fullpath = strdup(partpath);
  }

  return fullpath;
}

/*
** .......... SIGNAL HANDLERS ..........
*/
static void RequestStop(int signum)
{
  signal(signum,SIG_IGN);
  syslog(LOG_INFO,"Received request to shut down");
  Processing = 0;			/* WE WANT TO STOP */
  signal(signum,RequestStop);
}

static void RefreshConfig(int signum)
{
  signal(signum,SIG_IGN);
  syslog(LOG_INFO,"Received request to load new configuration");
  ReloadConfig = 1;
  signal(signum,RefreshConfig);
}
